/*
** This file is rebuilt and substitited every time you run a build.
** Things which need to be per-build should be defined here, declared
** in the version.h header, and then used wherever you want
*/
#include "version.h"

namespace Stochas
{
   const char* Build::MajorVersionStr = "1";
   const int   Build::MajorVersionInt = 1;
   
   const char* Build::SubVersionStr = "3";
   const int   Build::SubVersionInt = 3;
   
   const char* Build::ReleaseNumberStr = "1000";
   const char* Build::ReleaseStr = "makepkg";
   
   const char* Build::BuildNumberStr = "360d5ca"; // Build number to be sure that each result could identified.
   
   const char* Build::FullVersionStr = "1.3.10.makepkg.360d5ca";
   const char* Build::BuildHost = "arch-nspawn-237232";
   const char* Build::BuildArch = "x86_64";

   const char* Build::BuildLocation = "local";

   const char* Build::BuildDate = "2026-06-06";
   const char* Build::BuildTime = "19:38:23";

}

