/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace SeqImageX
{

//================== infobox.txt ==================
static const unsigned char temp_binary_data_6[] =
"Probabilistic Polyrhythmic Step Sequencer\n"
"\n"
"Copyright (C) 2020 Andrew Shakinovsky.\n"
"This program is free software: you can redistribute it and/or modify\n"
"it under the terms of the GNU General Public License as published by\n"
"the Free Software Foundation, either version 3 of the License, or\n"
"(at your option) any later version.\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program.  If not, see https://www.gnu.org/licenses/.\n"
"\n"
"WEBSITE:\n"
"https://stochas.org\n"
"CONTRIBUTORS:\n"
"Andrew Shakinovsky (@rudeog)\n"
"David Clissold - design, concepts\n"
"Paul Walker (@baconpaul) - continuous integration\n"
"John Valentine (@j5v) - logo design\n";

const char* infobox_txt = (const char*) temp_binary_data_6;
}
