/* =========================================================================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#pragma once

namespace SeqImageX
{
    extern const char*   exclamation64_png;
    const int            exclamation64_pngSize = 1407;

    extern const char*   not_png;
    const int            not_pngSize = 2567;

    extern const char*   play_png;
    const int            play_pngSize = 662;

    extern const char*   play2_png;
    const int            play2_pngSize = 3118;

    extern const char*   logo_no_bg_png;
    const int            logo_no_bg_pngSize = 35412;

    extern const char*   xmark464_png;
    const int            xmark464_pngSize = 1589;

    extern const char*   infobox_txt;
    const int            infobox_txtSize = 884;

    // Number of elements in the namedResourceList and originalFileNames arrays.
    const int namedResourceListSize = 7;

    // Points to the start of a list of resource names.
    extern const char* namedResourceList[];

    // Points to the start of a list of resource filenames.
    extern const char* originalFilenames[];

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding data and its size (or a null pointer if the name isn't found).
    const char* getNamedResource (const char* resourceNameUTF8, int& dataSizeInBytes);

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding original, non-mangled filename (or a null pointer if the name isn't found).
    const char* getNamedResourceOriginalFilename (const char* resourceNameUTF8);
}
